;
; PerlCRT.def
;
; Dll export file for PerlCRT.dll
; this is needed for GCC/Mingw32 builds of Perl, since GCC
; can't understand MSVC-ish .lib files
;
; Created from the output of 'nm PerlCRT.lib | grep "00000000 T"'
;       -- Benjamin Stuhl <sho_pi@hotmail.com> 10-17-1999

EXPORTS
  wscanf
  wprintf
  wctomb
  wcsxfrm
  wcstoul
  wcstombs
  wcstol
  wcstok
  wcstod
  wcsstr
  wcsspn
  wcsrchr
  wcspbrk
  wcsncpy
  wcsncmp
  wcsncat
  wcslen
  wcsftime
  wcscspn
  wcscpy
  wcscoll
  wcscmp
  wcschr
  wcscat
  vwprintf
  vswprintf
  vsprintf
  vprintf
  vfwprintf
  vfprintf
  ungetwc
  ungetc
  towupper
  towlower
  toupper
  tolower
  tmpnam
  tmpfile
  time
  tanh
  tan
  system
  swscanf
  swprintf
  strxfrm
  strtoul
  strtol
  strtok
  strtod
  strstr
  strspn
  strrchr
  strpbrk
  strncpy
  strncmp
  strncat
  strlen
  strftime
  strerror
  strcspn
  strcpy
  strcoll
  strcmp
  strchr
  strcat
  sscanf
  srand
  sqrt
  sprintf
  sinh
  sin
  signal
  setvbuf
  setlocale
  setbuf
  scanf
  rewind
  rename
  remove
  realloc
  rand
  raise
  qsort
  putwchar
  putwc
  puts
  putchar
  putc
  printf
  pow
  perror
  modf
  mktime
  memset
  memmove
  memcpy
  memcmp
  memchr
  mbtowc
  mbstowcs
  mblen
  malloc
  longjmp
  log10
  log
  localtime
  localeconv
  ldiv
  ldexp
  labs
  isxdigit
  iswxdigit
  iswupper
  iswspace
  iswpunct
  iswprint
  iswlower
  iswgraph
  iswdigit
  iswctype
  iswcntrl
  iswascii
  iswalpha
  iswalnum
  isupper
  isspace
  ispunct
  isprint
  islower
  isleadbyte
  isgraph
  isdigit
  iscntrl
  isalpha
  isalnum
  is_wctype
  gmtime
  getwchar
  getwc
  gets
  getenv
  getchar
  getc
  fwscanf
  fwrite
  fwprintf
  ftell
  fsetpos
  fseek
  fscanf
  frexp
  freopen
  free
  fread
  fputws
  fputwc
  fputs
  fputc
  fprintf
  fopen
  fmod
  floor
  fgetws
  fgetwc
  fgets
  fgetpos
  fgetc
  fflush
  ferror
  feof
  fclose
  fabs
  exp
  exit
  div
  difftime
  ctime
  cosh
  cos
  clock
  clearerr
  ceil
  calloc
  bsearch
  atol
  atoi
  atof
  atan2
  atan
  asin
  asctime
  acos
  abs
  abort
  _yn
  _y1
  _y0
  _wutime
  _wunlink
  _wtol
  _wtoi64
  _wtoi
  _wtmpnam
  _wtempnam
  _wsystem
  _wstrtime
  _wstrdate
  _wstati64
  _wstat
  _wsplitpath
  _wspawnvpe
  _wspawnvp
  _wspawnve
  _wspawnv
  _wspawnlpe
  _wspawnlp
  _wspawnle
  _wspawnl
  _wsopen
  _wsetlocale
  _wsearchenv
  _wrmdir
  _write
  _wrename
  _wremove
  _wputenv
  _wpopen
  _wperror
  _wopen
  _wmktemp
  _wmkdir
  _wmakepath
  _wgetenv
  _wgetdcwd
  _wgetcwd
  _wfullpath
  _wfsopen
  _wfreopen
  _wfopen
  _wfindnexti64
  _wfindnext
  _wfindfirsti64
  _wfindfirst
  _wfdopen
  _wexecvpe
  _wexecvp
  _wexecve
  _wexecv
  _wexeclpe
  _wexeclp
  _wexecle
  _wexecl
  _wctime
  _wcsupr
  _wcsset
  _wcsrev
  _wcsnset
  _wcsnicoll
  _wcsnicmp
  _wcsncoll
  _wcslwr
  _wcsicoll
  _wcsicmp
  _wcsdup
  _wcreat
  _wchmod
  _wchdir
  _wasctime
  _waccess
  _vsnwprintf
  _vsnprintf
  _utime
  _unlock
  _unloaddll
  _unlink
  _ungetch
  _umask
  _ultow
  _ultoa
  _ui64tow
  _ui64toa
  _tzset
  _toupper
  _tolower
  _tempnam
  _telli64
  _tell
  _swab
  _sys_nerr
  _strupr
  _strtime
  _strset
  _strrev
  _strnset
  _strnicoll
  _strnicmp
  _strncoll
  _strlwr
  _stricoll
  _stricmp
  _strerror
  _strdup
  _strdate
  _strcmpi
  _statusfp
  _stati64
  _stat
  _splitpath
  _spawnvpe
  _spawnvp
  _spawnve
  _spawnv
  _spawnlpe
  _spawnlp
  _spawnle
  _spawnl
  _sopen
  _snwprintf
  _snprintf
  _sleep
  _setsystime
  _setmode
  _setmbcp
  _setmaxstdio
  _setjmp3
  _setjmp
  _seterrormode
  _set_sbh_threshold
  _set_error_mode
  _seh_longjmp_unwind@4
  _searchenv
  _scalb
  _safe_fprem1
  _safe_fprem
  _safe_fdivr
  _safe_fdiv
  _rotr
  _rotl
  _rmtmp
  _rmdir
  _read
  _putws
  _putw
  _putenv
  _putch
  _purecall
  _popen
  _pipe
  _pclose
  _outpw
  _outpd
  _outp
  _open_osfhandle
  _open
  _nextafter
  _msize
  _mktemp
  _mkdir
  _memicmp
  _memccpy
  _mbsupr
  _mbstrlen
  _mbstok
  _mbsstr
  _mbsspnp
  _mbsspn
  _mbsset
  _mbsrev
  _mbsrchr
  _mbspbrk
  _mbsnset
  _mbsninc
  _mbsnicoll
  _mbsnicmp
  _mbsnextc
  _mbsncpy
  _mbsncoll
  _mbsncmp
  _mbsnccnt
  _mbsncat
  _mbsnbset
  _mbsnbicoll
  _mbsnbicmp
  _mbsnbcpy
  _mbsnbcoll
  _mbsnbcnt
  _mbsnbcmp
  _mbsnbcat
  _mbslwr
  _mbslen
  _mbsinc
  _mbsicoll
  _mbsicmp
  _mbsdup
  _mbsdec
  _mbscspn
  _mbscpy
  _mbscoll
  _mbscmp
  _mbschr
  _mbscat
  _mbsbtype
  _mbctoupper
  _mbctombb
  _mbctolower
  _mbctokata
  _mbctohira
  _mbclen
  _mbcjmstojis
  _mbcjistojms
  _mbccpy
  _mbbtype
  _mbbtombc
  _makepath
  _ltow
  _ltoa
  _lseeki64
  _lseek
  _lsearch
  _lrotr
  _lrotl
  _longjmpex
  _logb
  _locking
  _lock
  _local_unwind2
  _loaddll
  _lfind
  _kbhit
  _jn
  _j1
  _j0
  _itow
  _itoa
  _isnan
  _ismbstrail
  _ismbslead
  _ismbcupper
  _ismbcsymbol
  _ismbcspace
  _ismbcpunct
  _ismbcprint
  _ismbclower
  _ismbclegal
  _ismbcl2
  _ismbcl1
  _ismbcl0
  _ismbckata
  _ismbchira
  _ismbcgraph
  _ismbcdigit
  _ismbcalpha
  _ismbcalnum
  _ismbbtrail
  _ismbbpunct
  _ismbbprint
  _ismbblead
  _ismbbkpunct
  _ismbbkprint
  _ismbbkana
  _ismbbkalnum
  _ismbbgraph
  _ismbbalpha
  _ismbbalnum
  _isctype
  _isatty
  _inpw
  _inpd
  _inp
  _initterm
  _iob
  _i64tow
  _i64toa
  _hypot
  _HUGE
  _heapwalk
  _heapused
  _heapset
  _heapmin
  _heapchk
  _heapadd
  _global_unwind2
  _getws
  _getw
  _getsystime
  _getpid
  _getmbcp
  _getmaxstdio
  _getdrives
  _getdrive
  _getdllprocaddr
  _getdiskfree
  _getdcwd
  _getcwd
  _getche
  _getch
  _get_sbh_threshold
  _get_osfhandle
  _gcvt
  _futime
  _fullpath
  _ftol
  _ftime
  _fstati64
  _fstat
  _fsopen
  _free_osfhnd
  _fputwchar
  _fputchar
  _fpreset
  _fpieee_flt
  _fpclass
  _fmode
  _flushall
  _flsbuf
  _finite
  _findnexti64
  _findnext
  _findfirsti64
  _findfirst
  _findclose
  _fileno
  _filelengthi64
  _filelength
  _filbuf
  _fgetwchar
  _fgetchar
  _fdopen
  _fcvt
  _fcloseall
  _expand
  _exit
  _execvpe
  _execvp
  _execve
  _execv
  _execlpe
  _execlp
  _execle
  _execl
  _except_handler3
  _except_handler2
  _errno
  _eof
  _endthreadex
  _endthread
  _ecvt
  _dup2
  _dup
  _cwait
  _cscanf
  _creat
  _cputs
  _cprintf
  _copysign
  _controlfp
  _control87
  _commit
  _close
  _clearfp
  _chsize
  _chmod
  _chgsign
  _chdrive
  _chdir
  _cgets
  _cexit
  _callnewh
  _cabs
  _c_exit
  _beginthreadex
  _beginthread
  _beep
  _atoldbl
  _atoi64
  _atodbl
  _assert
  _amsg_exit
  _adj_fptan
  _adj_fprem1
  _adj_fprem
  _adj_fpatan
  _adj_fdivr_m64
  _adj_fdivr_m32i
  _adj_fdivr_m32
  _adj_fdivr_m16i
  _adj_fdiv_r
  _adj_fdiv_m64
  _adj_fdiv_m32i
  _adj_fdiv_m32
  _adj_fdiv_m16i
  _access
  _abnormal_termination
  __wgetmainargs
  __unDName
  __toascii
  __threadid
  __threadhandle
  __setusermatherr
  __set_app_type
  __pxcptinfoptrs
  __p__wpgmptr
  __p__winver
  __p__winminor
  __p__winmajor
  __p__wenviron
  __p__wcmdln
  __p__tzname
  __p__timezone
  __p__pwctype
  __p__pgmptr
  __p__pctype
  __p__osver
  __p__mbctype
  __p__mbcasemap
  __p__iob
  __p__fmode
  __p__fileinfo
  __p__environ
  __p__dstbias
  __p__daylight
  __p__commode
  __p__amblksiz
  __p__acmdln
  __p___winitenv
  __p___wargv
  __p___mb_cur_max
  __p___initenv
  __p___argv
  __p___argc
  __lconv_init
  __iscsymf
  __iscsym
  __isascii
  __getmainargs
  __fpecode
  __doserrno
  __dllonexit
  __crtLCMapStringA
  __crtGetLocaleInfoW
  __crtCompareStringA
  __STRINGTOLD
  __RTtypeid
  __RTDynamicCast
  __RTCastToVoid
  __CxxLongjmpUnwind@4
  __CxxFrameHandler
  _XcptFilter
  _Strftime
  _Gettnames
  _Getmonths
  _Getdays
  _EH_prolog
  _CxxThrowException@8
  _CItanh
  _CItan
  _CIsqrt
  _CIsinh
  _CIsin
  _CIpow
  _CIlog10
  _CIlog
  _CIfmod
  _CIexp
  _CIcosh
  _CIcos
  _CIatan2
  _CIatan
  _CIasin
  _CIacos
  $I10_OUTPUT
  _aullshr
  _aullrem
  _aulldiv
  _allshr
  _allshl
  _allrem
  _allmul
  _alldiv
  _setdefaultprecision
  _wsetargv
  _matherr
  _setargv
  __setargv
  _CRT_INIT@12
  _DllMainCRTStartup@12
  _onexit
  atexit
  _alloca_probe
  _chkstk
