;
; oldnames.def
;
; oldnames versions of MSVCRT/PerlCRT functions for GCC/Mingw32
;
; This file is taken from the Mingw32 distribution
;  Created by Colin Peters for Mingw32
;
; Added to Perl5 distrbution by Benjamin Stuhl <sho_pi@hotmail.com>

EXPORTS
access
beep
cabs
chdir
chmod
chsize
close
creat
cwait
dup
dup2
ecvt
eof
execl
execle
execlp
execlpe
execv
execve
execvp
execvpe
fcvt
fdopen
fgetchar
fgetwchar
filelength
fileno
fputchar
fputwchar
fstat
ftime
gcvt
getch
getche
getcwd
getpid
getw
heapwalk
hypot
isatty
itoa
j0
j1
jn
kbhit
lseek
ltoa
memccpy
memicmp
mkdir
mktemp
open
pclose
popen
putch
putenv
putw
read
rmdir
searchenv
seterrormode
setmode
sleep
sopen
spawnl
spawnle
spawnlp
spawnlpe
spawnv
spawnve
spawnvp
spawnvpe
stat
strcmpi
strdup
stricmp
stricoll
strlwr
strnicmp
strnset
strrev
strset
strupr
swab
tell
tempnam
tzset
umask
ungetch
unlink
utime
wcsdup
wcsicmp
wcsicoll
wcslwr
wcsnicmp
wcsnset
wcsrev
wcsset
wcsupr
write
y0
y1
yn
